@@ -1,4 +1,6 @@
 # coding: utf-8
+import inspect
+from typing import Any, List, Dict, Optional, TypeVar, Text, Generic
 
 from . import registry
 from .exceptions import (
@@ -14,6 +16,21 @@ from .exceptions import (
 from .utils import ugettext as _
 
 
+V = TypeVar('V')
+
+
+def is_callable_with_kwargs(f):
+    # python 3 variant -> return inspect.getfullargspec(f).varkw is not None
+    return inspect.getargspec(f).keywords is not None
+
+
+def call_with_args(f, *args, **kwargs):
+    if is_callable_with_kwargs(f):
+        f(*args, **kwargs)
+    else:
+        f()
+
+
 class CallableInstance(object):
     """
     Proxy that can override params by passing in kwargs and can run a callable.
@@ -75,10 +92,6 @@ class Transition(object):
     def __or__(self, other):
         return CombinedTransition(self, other, identifier=self.identifier)
 
-    def __contribute_to_class__(self, managed, identifier):
-        self.managed = managed
-        self.identifier = identifier
-
     def __get__(self, machine, owner):
         def transition_callback(*args, **kwargs):
             return self._run(machine, *args, **kwargs)
@@ -94,6 +107,9 @@ class Transition(object):
         self.on_execute = f
         return self
 
+    def _set_identifier(self, identifier):
+        self.identifier = identifier
+
     def _can_run(self, machine):
         if machine.current_state == self.source:
             return self
@@ -132,7 +148,12 @@ class Transition(object):
             if num < 2:
                 return result, destination
 
-            if isinstance(result[-1], State):
+            try:
+                last_item = result[-1]
+            except (AttributeError, TypeError, IndexError, KeyError):
+                last_item = None
+
+            if isinstance(last_item, State):
                 result, destination = result[:-1], result[-1]
                 if len(result) == 1:
                     result = result[0]
@@ -175,10 +196,10 @@ class CombinedTransition(Transition):
         self._right.on_execute = self.on_execute
         return result
 
-    def __contribute_to_class__(self, managed, identifier):
-        super(CombinedTransition, self).__contribute_to_class__(managed, identifier)
-        self._left.__contribute_to_class__(managed, identifier)
-        self._right.__contribute_to_class__(managed, identifier)
+    def _set_identifier(self, identifier):
+        super(CombinedTransition, self)._set_identifier(identifier)
+        self._left._set_identifier(identifier)
+        self._right._set_identifier(identifier)
 
     def _can_run(self, machine):
         return self._left._can_run(machine) or self._right._can_run(machine)
@@ -191,46 +212,74 @@ class CombinedTransition(Transition):
 
 class State(object):
 
-    def __init__(self, name, value=None, initial=False):
+    def __init__(self, name, value=None, initial=False, final=False):
+        # type: (Text, Optional[V], bool, bool) -> None
         self.name = name
-        self.identifier = None
         self.value = value
         self._initial = initial
-        self.transitions = []
+        self.identifier = None  # type: Optional[Text]
+        self.transitions = []  # type: List[Transition]
+        self._final = final
 
     def __repr__(self):
-        return "{}({!r}, identifier={!r}, value={!r}, initial={!r})".format(
-            type(self).__name__, self.name, self.identifier, self.value, self.initial)
+        return "{}({!r}, identifier={!r}, value={!r}, initial={!r}, final={!r})".format(
+            type(self).__name__, self.name, self.identifier, self.value, self.initial, self.final
+        )
+
+    def _set_identifier(self, identifier):
+        self.identifier = identifier
+        if self.value is None:
+            self.value = identifier
+
+    def _to_(self, *states):
+        transition = Transition(self, *states)
+        self.transitions.append(transition)
+        return transition
+
+    def _from_(self, *states):
+        combined = None
+        for origin in states:
+            transition = Transition(origin, self)
+            origin.transitions.append(transition)
+            if combined is None:
+                combined = transition
+            else:
+                combined |= transition
+        return combined
+
+    def _get_proxy_method_to_itself(self, method):
+        def proxy(*states):
+            return method(*states)
+
+        def proxy_to_itself():
+            return proxy(self)
+
+        proxy.itself = proxy_to_itself
+        return proxy
 
     @property
     def to(self):
-        def to_method(*states):
-            transition = Transition(self, *states)
-            self.transitions.append(transition)
-            return transition
+        return self._get_proxy_method_to_itself(self._to_)
 
-        def to_itself():
-            return to_method(self)
-
-        to_method.itself = to_itself
-        return to_method
-
-    def __contribute_to_class__(self, managed, identifier):
-        self.managed = managed
-        self.identifier = identifier
-        if not self.value:
-            self.value = identifier
+    @property
+    def from_(self):
+        return self._get_proxy_method_to_itself(self._from_)
 
     @property
     def initial(self):
         return self._initial
 
+    @property
+    def final(self):
+        return self._final
+
 
 def check_state_factory(state):
     "Return a property that checks if the current state is the desired state"
     @property
     def is_in_state(self):
-        return self.current_state == state
+        # type: (BaseStateMachine) -> bool
+        return bool(self.current_state == state)
     return is_in_state
 
 
@@ -241,22 +290,41 @@ class StateMachineMetaclass(type):
         registry.register(cls)
         cls.states = []
         cls.transitions = []
-        for key, value in sorted(attrs.items(), key=lambda v: v[0]):
-            if isinstance(value, State):
-                value.__contribute_to_class__(cls, key)
-                cls.states.append(value)
-            elif isinstance(value, Transition):
-                value.__contribute_to_class__(cls, key)
-                cls.transitions.append(value)
+        cls.states_map = {}
+        cls.add_inherited(bases)
+        cls.add_from_attributes(attrs)
 
         for state in cls.states:
             setattr(cls, 'is_{}'.format(state.identifier), check_state_factory(state))
 
-        cls.states_map = {s.value: s for s in cls.states}
+    def add_inherited(cls, bases):
+        for base in bases:
+            for state in getattr(base, 'states', []):
+                cls.add_state(state.identifier, state)
+            for transition in getattr(base, 'transitions', []):
+                cls.add_transition(transition.identifier, transition)
+
+    def add_from_attributes(cls, attrs):
+        for key, value in sorted(attrs.items(), key=lambda pair: pair[0]):
+            if isinstance(value, State):
+                cls.add_state(key, value)
+            elif isinstance(value, Transition):
+                cls.add_transition(key, value)
+
+    def add_state(cls, identifier, state):
+        state._set_identifier(identifier)
+        cls.states.append(state)
+        cls.states_map[state.value] = state
+
+    def add_transition(cls, identifier, transition):
+        transition._set_identifier(identifier)
+        cls.transitions.append(transition)
 
 
-class Model(object):
-    state = None
+class Model(Generic[V]):
+
+    def __init__(self):
+        self.state = None  # type: Optional[V]
 
     def __repr__(self):
         return 'Model(state={})'.format(self.state)
@@ -264,7 +332,12 @@ class Model(object):
 
 class BaseStateMachine(object):
 
+    transitions = []  # type: List[Transition]
+    states = []  # type: List[State]
+    states_map = {}  # type: Dict[Any, State]
+
     def __init__(self, model=None, state_field='state', start_value=None):
+        # type: (Any, str, Optional[V]) -> None
         self.model = model if model else Model()
         self.state_field = state_field
         self.start_value = start_value
@@ -277,6 +350,23 @@ class BaseStateMachine(object):
             self.current_state.identifier,
         )
 
+    def _visit_neighbour_states(self, transition, visited_states):
+        for neighbour_state in transition.destinations:
+            if neighbour_state not in visited_states:
+                self._visitable_states(neighbour_state, visited_states)
+
+    def _visitable_states(self, start_state, visited_states):
+        visited_states.append(start_state)
+
+        for transition in start_state.transitions:
+            self._visit_neighbour_states(transition, visited_states)
+        return visited_states
+
+    def _disconnected_states(self, starting_state):
+        visitable_states = self._visitable_states(starting_state, visited_states=[])
+
+        return set(self.states) - set(visitable_states)
+
     def check(self):
         if not self.states:
             raise InvalidDefinition(_('There are no states.'))
@@ -290,26 +380,52 @@ class BaseStateMachine(object):
                                       'Your currently have these: {!r}'.format(initials)))
         self.initial_state = initials[0]
 
+        disconnected_states = self._disconnected_states(self.initial_state)
+        if (disconnected_states):
+            raise InvalidDefinition(_('There are unreachable states. '
+                                    'The statemachine graph should have a single component. '
+                                      'Disconnected states: [{}]'.format(disconnected_states)))
+
+        final_state_with_invalid_transitions = [
+            state for state in self.final_states if state.transitions
+        ]
+
+        if final_state_with_invalid_transitions:
+            raise InvalidDefinition(_('Final state does not should have defined '
+                                      'transitions starting from that state'))
+
         if self.current_state_value is None:
             if self.start_value:
                 self.current_state_value = self.start_value
             else:
                 self.current_state_value = self.initial_state.value
 
+    @property
+    def final_states(self):
+        return [state for state in self.states if state.final]
+
     @property
     def current_state_value(self):
-        return getattr(self.model, self.state_field, None)
+        # type: () -> V
+        value = getattr(self.model, self.state_field, None)  # type: V
+        return value
 
     @current_state_value.setter
     def current_state_value(self, value):
+        # type: (V) -> None
         if value not in self.states_map:
             raise InvalidStateValue(value)
         setattr(self.model, self.state_field, value)
 
     @property
     def current_state(self):
+        # type: () -> State
         return self.states_map[self.current_state_value]
 
+    @current_state.setter
+    def current_state(self, value):
+        self.current_state_value = value.value
+
     @property
     def allowed_transitions(self):
         "get the callable proxy of the current allowed transitions"
@@ -318,10 +434,6 @@ class BaseStateMachine(object):
             for t in self.current_state.transitions if t._can_run(self)
         ]
 
-    @current_state.setter
-    def current_state(self, value):
-        self.current_state_value = value.value
-
     def _activate(self, transition, *args, **kwargs):
         bounded_on_event = getattr(self, 'on_{}'.format(transition.identifier), None)
         on_event = transition.on_execute
@@ -341,25 +453,28 @@ class BaseStateMachine(object):
         if callable(bounded_on_exit_state_event):
             bounded_on_exit_state_event(self.current_state)
 
+        bounded_on_exit_specific_state_event = getattr(
+            self, 'on_exit_{}'.format(self.current_state.identifier), None)
+
+        if callable(bounded_on_exit_specific_state_event):
+            call_with_args(bounded_on_exit_specific_state_event, *args, **kwargs)
+
+        self.current_state = destination
+
         bounded_on_enter_state_event = getattr(self, 'on_enter_state', None)
         if callable(bounded_on_enter_state_event):
             bounded_on_enter_state_event(destination)
 
-        bounded_on_exit_specific_state_event = getattr(
-            self, 'on_exit_{}'.format(self.current_state.identifier), None)
-        if callable(bounded_on_exit_specific_state_event):
-            bounded_on_exit_specific_state_event()
-
         bounded_on_enter_specific_state_event = getattr(
             self, 'on_enter_{}'.format(destination.identifier), None)
         if callable(bounded_on_enter_specific_state_event):
-            bounded_on_enter_specific_state_event()
+            call_with_args(bounded_on_enter_specific_state_event, *args, **kwargs)
 
-        self.current_state = destination
         return result
 
     def get_transition(self, transition_identifier):
-        transition = getattr(self, transition_identifier, None)
+        # type: (Text) -> CallableInstance
+        transition = getattr(self, transition_identifier, None)  # type: CallableInstance
         if not hasattr(transition, 'source') or not callable(transition):
             raise InvalidTransitionIdentifier(transition_identifier)
         return transition
