@@ -50,6 +50,14 @@ def test_machine_should_only_allow_only_one_initial_state():
         CampaignMachine(model)
 
 
+def test_machine_should_not_allow_transitions_from_final_state(
+    campaign_machine_with_invalid_final_state_transition
+):
+    with pytest.raises(exceptions.InvalidDefinition):
+        model = MyModel()
+        campaign_machine_with_invalid_final_state_transition(model)
+
+
 def test_should_change_state(campaign_machine):
     model = MyModel()
     machine = campaign_machine(model)
@@ -342,3 +350,26 @@ def test_should_not_create_big_disconnected_machine():
         assert 'Magenta' in e.message
         assert 'Blue' in e.message
         assert 'Cyan' not in e.message
+
+
+def test_state_value_is_correct():
+    STATE_NEW = 0
+    STATE_DRAFT = 1
+
+    class ValueTestModel(StateMachine):
+        new = State(STATE_NEW, value=STATE_NEW, initial=True)
+        draft = State(STATE_DRAFT, value=STATE_DRAFT)
+
+        write = new.to(draft)
+
+    model = ValueTestModel()
+    assert model.new.value == STATE_NEW
+    assert model.draft.value == STATE_DRAFT
+
+
+def test_final_states(campaign_machine_with_final_state):
+    model = MyModel()
+    machine = campaign_machine_with_final_state(model)
+    final_states = machine.final_states
+    assert len(final_states) == 1
+    assert final_states[0].name == "Closed"
